@@ -1,4 +1,4 @@
-// Copyright 2021 Colin Finck <colin@reactos.org>
+// Copyright 2021-2022 Colin Finck <colin@reactos.org>
 // SPDX-License-Identifier: MIT OR Apache-2.0
 
 mod sector_reader;
@@ -9,13 +9,15 @@ use std::io;
 use std::io::{BufReader, Read, Seek, Write};
 
 use anyhow::{anyhow, bail, Context, Result};
-use chrono::{DateTime, Utc};
 use ntfs::attribute_value::NtfsAttributeValue;
 use ntfs::indexes::NtfsFileNameIndex;
 use ntfs::structured_values::{
     NtfsAttributeList, NtfsFileName, NtfsFileNamespace, NtfsStandardInformation,
 };
 use ntfs::{Ntfs, NtfsAttribute, NtfsAttributeType, NtfsFile, NtfsReadSeek};
+use time::format_description::FormatItem;
+use time::macros::format_description;
+use time::OffsetDateTime;
 
 use sector_reader::SectorReader;
 
@@ -41,7 +43,7 @@ fn main() -> Result<()> {
     }
 
     let f = File::open(&args[1])?;
-    let sr = SectorReader::new(f, 512)?;
+    let sr = SectorReader::new(f, 4096)?;
     let mut fs = BufReader::new(sr);
     let mut ntfs = Ntfs::new(&mut fs)?;
     ntfs.read_upcase_table(&mut fs)?;
@@ -120,6 +122,7 @@ where
         let ty = attribute.ty()?;
 
         attr_print_attribute(
+            info,
             with_runs,
             &attribute,
             file.file_record_number(),
@@ -143,6 +146,7 @@ where
                 let entry_attribute = entry.to_attribute(&entry_file)?;
 
                 attr_print_attribute(
+                    info,
                     with_runs,
                     &entry_attribute,
                     entry_record_number,
@@ -156,13 +160,17 @@ where
     Ok(())
 }
 
-fn attr_print_attribute<'n>(
+fn attr_print_attribute<'n, T>(
+    info: &mut CommandInfo<T>,
     with_runs: bool,
     attribute: &NtfsAttribute<'n, '_>,
     record_number: u64,
     attribute_prefix: &str,
     data_run_prefix: &str,
-) -> Result<()> {
+) -> Result<()>
+where
+    T: Read + Seek,
+{
     let instance = format!("{}{}", attribute_prefix, attribute.instance());
     let ty = attribute.ty()?;
     let resident = attribute.is_resident();
@@ -176,11 +184,13 @@ fn attr_print_attribute<'n>(
     );
 
     if with_runs {
-        if let NtfsAttributeValue::NonResident(non_resident_value) = attribute.value()? {
+        let value = attribute.value(&mut info.fs)?;
+
+        if let NtfsAttributeValue::NonResident(non_resident_value) = value {
             for (i, data_run) in non_resident_value.data_runs().enumerate() {
                 let data_run = data_run?;
                 let instance = format!("{}{}", data_run_prefix, i);
-                let start = data_run.data_position().unwrap_or(0);
+                let start = data_run.data_position();
                 let length = data_run.allocated_size();
 
                 println!(
@@ -344,6 +354,9 @@ where
 }
 
 fn fileinfo_std(attribute: NtfsAttribute) -> Result<()> {
+    const TIME_FORMAT: &[FormatItem] =
+        format_description!("[year]-[month]-[day] [hour]:[minute]:[second] UTC");
+
     println!();
     println!("{:=^72}", " STANDARD INFORMATION ");
 
@@ -351,11 +364,18 @@ fn fileinfo_std(attribute: NtfsAttribute) -> Result<()> {
 
     println!("{:34}{:?}", "Attributes:", std_info.file_attributes());
 
-    let format = "%F %T UTC";
-    let atime = DateTime::<Utc>::from(std_info.access_time()).format(format);
-    let ctime = DateTime::<Utc>::from(std_info.creation_time()).format(format);
-    let mtime = DateTime::<Utc>::from(std_info.modification_time()).format(format);
-    let mmtime = DateTime::<Utc>::from(std_info.mft_record_modification_time()).format(format);
+    let atime = OffsetDateTime::from(std_info.access_time())
+        .format(TIME_FORMAT)
+        .unwrap();
+    let ctime = OffsetDateTime::from(std_info.creation_time())
+        .format(TIME_FORMAT)
+        .unwrap();
+    let mtime = OffsetDateTime::from(std_info.modification_time())
+        .format(TIME_FORMAT)
+        .unwrap();
+    let mmtime = OffsetDateTime::from(std_info.mft_record_modification_time())
+        .format(TIME_FORMAT)
+        .unwrap();
     println!("{:34}{}", "Access Time:", atime);
     println!("{:34}{}", "Creation Time:", ctime);
     println!("{:34}{}", "Modification Time:", mtime);
@@ -476,13 +496,13 @@ where
     let output_file_name = if data_stream_name.is_empty() {
         file_name.to_string()
     } else {
-        format!("{}_{}", file_name, data_stream_name)
+        format!("{file_name}_{data_stream_name}")
     };
     let mut output_file = OpenOptions::new()
         .write(true)
         .create_new(true)
         .open(&output_file_name)
-        .with_context(|| format!("Tried to open \"{}\" for writing", output_file_name))?;
+        .with_context(|| format!("Tried to open \"{output_file_name}\" for writing"))?;
 
     // Open the desired file and find the $DATA attribute we are looking for.
     let file = parse_file_arg(file_name, info)?;
@@ -498,7 +518,7 @@ where
     };
     let data_item = data_item?;
     let data_attribute = data_item.to_attribute();
-    let mut data_value = data_attribute.value()?;
+    let mut data_value = data_attribute.value(&mut info.fs)?;
 
     println!(
         "Saving {} bytes of data in \"{}\"...",
