@@ -1,6 +1,17 @@
-// Copyright 2021 Colin Finck <colin@reactos.org>
+// Copyright 2021-2022 Colin Finck <colin@reactos.org>
 // SPDX-License-Identifier: MIT OR Apache-2.0
 
+use core::iter::FusedIterator;
+use core::mem;
+use core::ops::Range;
+
+use binread::io::{Read, Seek};
+use bitflags::bitflags;
+use byteorder::{ByteOrder, LittleEndian};
+use enumn::N;
+use memoffset::offset_of;
+use strum_macros::Display;
+
 use crate::attribute_value::{
     NtfsAttributeListNonResidentAttributeValue, NtfsAttributeValue, NtfsNonResidentAttributeValue,
     NtfsResidentAttributeValue,
@@ -12,16 +23,7 @@ use crate::structured_values::{
     NtfsAttributeList, NtfsAttributeListEntries, NtfsStructuredValue,
     NtfsStructuredValueFromResidentAttributeValue,
 };
-use crate::types::Vcn;
-use binread::io::{Read, Seek};
-use bitflags::bitflags;
-use byteorder::{ByteOrder, LittleEndian};
-use core::iter::FusedIterator;
-use core::mem;
-use core::ops::Range;
-use enumn::N;
-use memoffset::offset_of;
-use strum_macros::Display;
+use crate::types::{NtfsPosition, Vcn};
 
 /// On-disk structure of the generic header of an NTFS Attribute.
 #[repr(C, packed)]
@@ -196,6 +198,19 @@ impl<'n, 'f> NtfsAttribute<'n, 'f> {
         LittleEndian::read_u32(&self.file.record_data()[start..])
     }
 
+    pub(crate) fn ensure_ty(&self, expected: NtfsAttributeType) -> Result<()> {
+        let ty = self.ty()?;
+        if ty != expected {
+            return Err(NtfsError::AttributeOfDifferentType {
+                position: self.position(),
+                expected,
+                actual: ty,
+            });
+        }
+
+        Ok(())
+    }
+
     /// Returns flags set for this attribute as specified by [`NtfsAttributeFlags`].
     pub fn flags(&self) -> NtfsAttributeFlags {
         let start = self.offset + offset_of!(NtfsAttributeHeader, flags);
@@ -263,12 +278,12 @@ impl<'n, 'f> NtfsAttribute<'n, 'f> {
         )
     }
 
-    pub(crate) fn non_resident_value_data_and_position(&self) -> (&'f [u8], u64) {
+    pub(crate) fn non_resident_value_data_and_position(&self) -> (&'f [u8], NtfsPosition) {
         debug_assert!(!self.is_resident());
         let start = self.offset + self.non_resident_value_data_runs_offset() as usize;
         let end = self.offset + self.attribute_length() as usize;
         let data = &self.file.record_data()[start..end];
-        let position = self.file.position() + start as u64;
+        let position = self.file.position() + start;
 
         (data, position)
     }
@@ -290,8 +305,8 @@ impl<'n, 'f> NtfsAttribute<'n, 'f> {
     }
 
     /// Returns the absolute position of this NTFS Attribute within the filesystem, in bytes.
-    pub fn position(&self) -> u64 {
-        self.file.position() + self.offset as u64
+    pub fn position(&self) -> NtfsPosition {
+        self.file.position() + self.offset
     }
 
     /// Attempts to parse the value data as the given resident structured value type and returns that.
@@ -307,14 +322,7 @@ impl<'n, 'f> NtfsAttribute<'n, 'f> {
     where
         S: NtfsStructuredValueFromResidentAttributeValue<'n, 'f>,
     {
-        let ty = self.ty()?;
-        if ty != S::TY {
-            return Err(NtfsError::AttributeOfDifferentType {
-                position: self.position(),
-                expected: S::TY,
-                actual: ty,
-            });
-        }
+        self.ensure_ty(S::TY)?;
 
         if !self.is_resident() {
             return Err(NtfsError::UnexpectedNonResidentAttribute {
@@ -359,16 +367,9 @@ impl<'n, 'f> NtfsAttribute<'n, 'f> {
         T: Read + Seek,
         S: NtfsStructuredValue<'n, 'f>,
     {
-        let ty = self.ty()?;
-        if ty != S::TY {
-            return Err(NtfsError::AttributeOfDifferentType {
-                position: self.position(),
-                expected: S::TY,
-                actual: ty,
-            });
-        }
-
-        S::from_attribute_value(fs, self.value()?)
+        self.ensure_ty(S::TY)?;
+        let value = self.value(fs)?;
+        S::from_attribute_value(fs, value)
     }
 
     /// Returns the type of this NTFS Attribute, or [`NtfsError::UnsupportedAttributeType`]
@@ -409,7 +410,7 @@ impl<'n, 'f> NtfsAttribute<'n, 'f> {
         debug_assert!(self.is_resident());
 
         let start = self.resident_value_offset();
-        if start as u32 >= self.attribute_length() {
+        if start as u32 > self.attribute_length() {
             return Err(NtfsError::InvalidResidentAttributeValueOffset {
                 position: self.position(),
                 expected: start,
@@ -430,7 +431,10 @@ impl<'n, 'f> NtfsAttribute<'n, 'f> {
     }
 
     /// Returns an [`NtfsAttributeValue`] structure to read the value of this NTFS Attribute.
-    pub fn value(&self) -> Result<NtfsAttributeValue<'n, 'f>> {
+    pub fn value<T>(&self, fs: &mut T) -> Result<NtfsAttributeValue<'n, 'f>>
+    where
+        T: Read + Seek,
+    {
         if let Some(list_entries) = self.list_entries {
             // The first attribute reports the entire data size for all connected attributes
             // (remaining ones are set to zero).
@@ -439,11 +443,12 @@ impl<'n, 'f> NtfsAttribute<'n, 'f> {
 
             let value = NtfsAttributeListNonResidentAttributeValue::new(
                 self.file.ntfs(),
+                fs,
                 list_entries.clone(),
                 self.instance(),
                 self.ty()?,
                 data_size,
-            );
+            )?;
             Ok(NtfsAttributeValue::AttributeListNonResident(value))
         } else if self.is_resident() {
             let value = self.resident_value()?;
@@ -701,3 +706,38 @@ impl<'n, 'f> Iterator for NtfsAttributesRaw<'n, 'f> {
 }
 
 impl<'n, 'f> FusedIterator for NtfsAttributesRaw<'n, 'f> {}
+
+#[cfg(test)]
+mod tests {
+    use crate::indexes::NtfsFileNameIndex;
+    use crate::ntfs::Ntfs;
+    use crate::traits::NtfsReadSeek;
+
+    #[test]
+    fn test_empty_data_attribute() {
+        let mut testfs1 = crate::helpers::tests::testfs1();
+        let mut ntfs = Ntfs::new(&mut testfs1).unwrap();
+        ntfs.read_upcase_table(&mut testfs1).unwrap();
+        let root_dir = ntfs.root_directory(&mut testfs1).unwrap();
+
+        // Find the "empty-file".
+        let root_dir_index = root_dir.directory_index(&mut testfs1).unwrap();
+        let mut root_dir_finder = root_dir_index.finder();
+        let entry =
+            NtfsFileNameIndex::find(&mut root_dir_finder, &ntfs, &mut testfs1, "empty-file")
+                .unwrap()
+                .unwrap();
+        let empty_file = entry.to_file(&ntfs, &mut testfs1).unwrap();
+
+        let data_attribute_item = empty_file.data(&mut testfs1, "").unwrap().unwrap();
+        let data_attribute = data_attribute_item.to_attribute();
+        assert_eq!(data_attribute.value_length(), 0);
+
+        let mut data_attribute_value = data_attribute.value(&mut testfs1).unwrap();
+        assert!(data_attribute_value.is_empty());
+
+        let mut buf = [0u8; 5];
+        let bytes_read = data_attribute_value.read(&mut testfs1, &mut buf).unwrap();
+        assert_eq!(bytes_read, 0);
+    }
+}
