@@ -1,11 +1,10 @@
-// Copyright 2021 Colin Finck <colin@reactos.org>
+// Copyright 2021-2022 Colin Finck <colin@reactos.org>
 // SPDX-License-Identifier: MIT OR Apache-2.0
 //
 //! This module implements a reader for a non-resident attribute value (that is not part of an Attribute List).
 //! Non-resident attribute values are split up into one or more data runs, which are spread across the filesystem.
 //! This reader provides one contiguous data stream for all data runs.
 
-use core::convert::TryFrom;
 use core::iter::FusedIterator;
 use core::mem;
 
@@ -18,7 +17,7 @@ use super::seek_contiguous;
 use crate::error::{NtfsError, Result};
 use crate::ntfs::Ntfs;
 use crate::traits::NtfsReadSeek;
-use crate::types::{Lcn, Vcn};
+use crate::types::{Lcn, NtfsPosition, Vcn};
 
 /// Reader for a non-resident attribute value (whose data is in a cluster range outside the File Record).
 #[derive(Clone, Debug)]
@@ -28,7 +27,7 @@ pub struct NtfsNonResidentAttributeValue<'n, 'f> {
     /// Attribute bytes where the Data Run information of this non-resident value is stored on the filesystem.
     data: &'f [u8],
     /// Absolute position of the Data Run information within the filesystem, in bytes.
-    position: u64,
+    position: NtfsPosition,
     /// Iterator of data runs used for reading/seeking.
     stream_data_runs: NtfsDataRuns<'n, 'f>,
     /// Iteration state of the current Data Run.
@@ -39,25 +38,22 @@ impl<'n, 'f> NtfsNonResidentAttributeValue<'n, 'f> {
     pub(crate) fn new(
         ntfs: &'n Ntfs,
         data: &'f [u8],
-        position: u64,
+        position: NtfsPosition,
         data_size: u64,
     ) -> Result<Self> {
-        let mut stream_data_runs = NtfsDataRuns::new(ntfs, data, position);
-        let mut stream_state = StreamState::new(data_size);
+        let stream_data_runs = NtfsDataRuns::new(ntfs, data, position);
+        let stream_state = StreamState::new(data_size);
 
-        // Get the first Data Run already here to let `data_position` return something meaningful.
-        if let Some(stream_data_run) = stream_data_runs.next() {
-            let stream_data_run = stream_data_run?;
-            stream_state.set_stream_data_run(stream_data_run);
-        }
-
-        Ok(Self {
+        let mut value = Self {
             ntfs,
             data,
             position,
             stream_data_runs,
             stream_state,
-        })
+        };
+        value.next_data_run()?;
+
+        Ok(value)
     }
 
     /// Returns a variant of this reader that implements [`Read`] and [`Seek`]
@@ -75,8 +71,9 @@ impl<'n, 'f> NtfsNonResidentAttributeValue<'n, 'f> {
     /// Returns the absolute current data seek position within the filesystem, in bytes.
     /// This may be `None` if:
     ///   * The current seek position is outside the valid range, or
+    ///   * The attribute does not have a Data Run, or
     ///   * The current Data Run is a "sparse" Data Run
-    pub fn data_position(&self) -> Option<u64> {
+    pub fn data_position(&self) -> NtfsPosition {
         self.stream_state.data_position()
     }
 
@@ -112,9 +109,13 @@ impl<'n, 'f> NtfsNonResidentAttributeValue<'n, 'f> {
         self.ntfs
     }
 
-    /// Returns the absolute position of the Data Run information within the filesystem, in bytes.
-    pub fn position(&self) -> u64 {
-        self.position
+    /// Rewinds this value reader to the very beginning.
+    fn rewind(&mut self) -> Result<()> {
+        self.stream_data_runs = self.data_runs();
+        self.stream_state = StreamState::new(self.len());
+        self.next_data_run()?;
+
+        Ok(())
     }
 }
 
@@ -153,9 +154,7 @@ impl<'n, 'f> NtfsReadSeek for NtfsNonResidentAttributeValue<'n, 'f> {
 
         let mut bytes_left_to_seek = match pos {
             SeekFrom::Start(n) => {
-                // Rewind to the very beginning.
-                self.stream_data_runs = self.data_runs();
-                self.stream_state = StreamState::new(self.len());
+                self.rewind()?;
                 n
             }
             SeekFrom::Current(n) if n >= 0 => n as u64,
@@ -217,8 +216,9 @@ where
     /// Returns the absolute current data seek position within the filesystem, in bytes.
     /// This may be `None` if:
     ///   * The current seek position is outside the valid range, or
+    ///   * The attribute does not have a Data Run, or
     ///   * The current Data Run is a "sparse" Data Run.
-    pub fn data_position(&self) -> Option<u64> {
+    pub fn data_position(&self) -> NtfsPosition {
         self.value.data_position()
     }
 
@@ -266,12 +266,12 @@ where
 pub struct NtfsDataRuns<'n, 'f> {
     ntfs: &'n Ntfs,
     data: &'f [u8],
-    position: u64,
+    position: NtfsPosition,
     state: DataRunsState,
 }
 
 impl<'n, 'f> NtfsDataRuns<'n, 'f> {
-    pub(crate) fn new(ntfs: &'n Ntfs, data: &'f [u8], position: u64) -> Self {
+    pub(crate) fn new(ntfs: &'n Ntfs, data: &'f [u8], position: NtfsPosition) -> Self {
         let state = DataRunsState {
             offset: 0,
             previous_lcn: Lcn::from(0),
@@ -288,7 +288,7 @@ impl<'n, 'f> NtfsDataRuns<'n, 'f> {
     pub(crate) fn from_state(
         ntfs: &'n Ntfs,
         data: &'f [u8],
-        position: u64,
+        position: NtfsPosition,
         state: DataRunsState,
     ) -> Self {
         Self {
@@ -304,8 +304,8 @@ impl<'n, 'f> NtfsDataRuns<'n, 'f> {
     }
 
     /// Returns the absolute position of the current Data Run header within the filesystem, in bytes.
-    pub fn position(&self) -> u64 {
-        self.position + self.state.offset as u64
+    pub fn position(&self) -> NtfsPosition {
+        self.position + self.state.offset
     }
 
     fn read_variable_length_bytes(
@@ -424,7 +424,7 @@ pub(crate) struct DataRunsState {
 pub struct NtfsDataRun {
     /// Absolute position of the Data Run within the filesystem, in bytes.
     /// This may be zero if this is a "sparse" Data Run.
-    position: u64,
+    position: NtfsPosition,
     /// Total allocated size of the Data Run, in bytes.
     /// The actual size used by data may be lower, but a Data Run does not know about that.
     allocated_size: u64,
@@ -455,11 +455,11 @@ impl NtfsDataRun {
     /// This may be `None` if:
     ///   * The current seek position is outside the valid range, or
     ///   * The Data Run is a "sparse" Data Run
-    pub fn data_position(&self) -> Option<u64> {
-        if self.position > 0 && self.stream_position < self.allocated_size() {
-            Some(self.position + self.stream_position)
+    pub fn data_position(&self) -> NtfsPosition {
+        if self.stream_position <= self.allocated_size() {
+            self.position + self.stream_position
         } else {
-            None
+            NtfsPosition::none()
         }
     }
 
@@ -480,15 +480,14 @@ impl NtfsReadSeek for NtfsDataRun {
         let bytes_to_read = usize::min(buf.len(), self.remaining_len() as usize);
         let work_slice = &mut buf[..bytes_to_read];
 
-        let bytes_read = if self.position == 0 {
+        let bytes_read = if let Some(position) = self.position.value() {
+            // This Data Run contains "real" data.
+            fs.seek(SeekFrom::Start(position.get() + self.stream_position))?;
+            fs.read(work_slice)?
+        } else {
             // This is a sparse Data Run.
             work_slice.fill(0);
             work_slice.len()
-        } else {
-            // This Data Run contains "real" data.
-            // We have already performed all necessary sanity checks above, so we can just unwrap here.
-            fs.seek(SeekFrom::Start(self.data_position().unwrap()))?;
-            fs.read(work_slice)?
         };
 
         self.stream_position += bytes_read as u64;
@@ -530,10 +529,14 @@ impl StreamState {
     /// Returns the absolute current data seek position within the filesystem, in bytes.
     /// This may be `None` if:
     ///   * The current seek position is outside the valid range, or
+    ///   * The attribute does not have a Data Run, or
     ///   * The current Data Run is a "sparse" Data Run
-    pub(crate) fn data_position(&self) -> Option<u64> {
-        let stream_data_run = self.stream_data_run.as_ref()?;
-        stream_data_run.data_position()
+    pub(crate) fn data_position(&self) -> NtfsPosition {
+        if let Some(stream_data_run) = self.stream_data_run.as_ref() {
+            stream_data_run.data_position()
+        } else {
+            NtfsPosition::none()
+        }
     }
 
     /// Returns the total (used) data size of the value, in bytes.
