@@ -1,6 +1,13 @@
-// Copyright 2021 Colin Finck <colin@reactos.org>
+// Copyright 2021-2022 Colin Finck <colin@reactos.org>
 // SPDX-License-Identifier: MIT OR Apache-2.0
 
+use core::num::NonZeroU64;
+
+use binread::io::{Read, Seek, SeekFrom};
+use bitflags::bitflags;
+use byteorder::{ByteOrder, LittleEndian};
+use memoffset::offset_of;
+
 use crate::attribute::{NtfsAttributeItem, NtfsAttributeType, NtfsAttributes, NtfsAttributesRaw};
 use crate::error::{NtfsError, Result};
 use crate::file_reference::NtfsFileReference;
@@ -12,10 +19,7 @@ use crate::structured_values::{
     NtfsFileName, NtfsFileNamespace, NtfsIndexRoot, NtfsStandardInformation,
     NtfsStructuredValueFromResidentAttributeValue,
 };
-use binread::io::{Read, Seek, SeekFrom};
-use bitflags::bitflags;
-use byteorder::{ByteOrder, LittleEndian};
-use memoffset::offset_of;
+use crate::types::NtfsPosition;
 
 /// A list of standardized NTFS File Record Numbers.
 ///
@@ -107,7 +111,8 @@ bitflags! {
 /// [`NtfsAttribute`]: crate::attribute::NtfsAttribute
 #[derive(Clone, Debug)]
 pub struct NtfsFile<'n> {
-    record: Record<'n>,
+    ntfs: &'n Ntfs,
+    record: Record,
     file_record_number: u64,
 }
 
@@ -115,21 +120,22 @@ impl<'n> NtfsFile<'n> {
     pub(crate) fn new<T>(
         ntfs: &'n Ntfs,
         fs: &mut T,
-        position: u64,
+        position: NonZeroU64,
         file_record_number: u64,
     ) -> Result<Self>
     where
         T: Read + Seek,
     {
         let mut data = vec![0; ntfs.file_record_size() as usize];
-        fs.seek(SeekFrom::Start(position))?;
+        fs.seek(SeekFrom::Start(position.get()))?;
         fs.read_exact(&mut data)?;
 
-        let mut record = Record::new(ntfs, data, position);
+        let mut record = Record::new(data, position.into());
         Self::validate_signature(&record)?;
         record.fixup()?;
 
         let file = Self {
+            ntfs,
             record,
             file_record_number,
         };
@@ -242,36 +248,23 @@ impl<'n> NtfsFile<'n> {
         let directory_index_name = "$I30";
 
         // The IndexRoot attribute is always resident and has to exist for every directory.
-        let index_root = self.find_resident_attribute_structured_value::<NtfsIndexRoot>(Some(
-            directory_index_name,
-        ))?;
+        let index_root_item =
+            self.find_attribute(fs, NtfsAttributeType::IndexRoot, Some(directory_index_name))?;
+        let index_root_attribute = index_root_item.to_attribute();
+        let index_root = index_root_attribute.resident_structured_value::<NtfsIndexRoot>()?;
 
         // The IndexAllocation attribute is only required for "large" indexes.
         // It is always non-resident and may even be in an Attribute List.
         let mut index_allocation_item = None;
         if index_root.is_large_index() {
-            let mut iter = self.attributes();
-
-            while let Some(item) = iter.next(fs) {
-                let item = item?;
-                let attribute = item.to_attribute();
-
-                let ty = attribute.ty()?;
-                if ty != NtfsAttributeType::IndexAllocation {
-                    continue;
-                }
-
-                let name = attribute.name()?;
-                if name != directory_index_name {
-                    continue;
-                }
-
-                index_allocation_item = Some(item);
-                break;
-            }
+            index_allocation_item = Some(self.find_attribute(
+                fs,
+                NtfsAttributeType::IndexAllocation,
+                Some(directory_index_name),
+            )?);
         }
 
-        NtfsIndex::<NtfsFileNameIndex>::new(index_root, index_allocation_item)
+        NtfsIndex::<NtfsFileNameIndex>::new(index_root_item, index_allocation_item)
     }
 
     /// Returns the NTFS File Record Number of this file.
@@ -282,10 +275,50 @@ impl<'n> NtfsFile<'n> {
         self.file_record_number
     }
 
+    /// Finds an attribute of a specific type, optionally with a specific name, and returns its [`NtfsAttributeItem`].
+    /// Returns [`NtfsError::AttributeNotFound`] if no such attribute could be found.
+    ///
+    /// This function also traverses Attribute Lists to find the attribute.
+    fn find_attribute<'f, T>(
+        &'f self,
+        fs: &mut T,
+        ty: NtfsAttributeType,
+        match_name: Option<&str>,
+    ) -> Result<NtfsAttributeItem<'n, 'f>>
+    where
+        T: Read + Seek,
+    {
+        let mut iter = self.attributes();
+
+        while let Some(item) = iter.next(fs) {
+            let item = item?;
+            let attribute = item.to_attribute();
+
+            if attribute.ty()? != ty {
+                continue;
+            }
+
+            if let Some(name) = match_name {
+                if attribute.name()? != name {
+                    continue;
+                }
+            }
+
+            return Ok(item);
+        }
+
+        Err(NtfsError::AttributeNotFound {
+            position: self.position(),
+            ty,
+        })
+    }
+
     /// Finds a resident attribute of a specific type, optionally with a specific name, and returns its structured value.
     /// Returns [`NtfsError::AttributeNotFound`] if no such resident attribute could be found.
     ///
     /// The attribute type is given through the passed structured value type parameter.
+    ///
+    /// Note that this function DOES NOT traverse Attribute Lists!
     pub(crate) fn find_resident_attribute_structured_value<'f, S>(
         &'f self,
         match_name: Option<&str>,
@@ -398,11 +431,11 @@ impl<'n> NtfsFile<'n> {
 
     /// Returns the [`Ntfs`] object reference associated to this file.
     pub fn ntfs(&self) -> &'n Ntfs {
-        self.record.ntfs()
+        self.ntfs
     }
 
     /// Returns the absolute byte position of this File Record in the NTFS filesystem.
-    pub fn position(&self) -> u64 {
+    pub fn position(&self) -> NtfsPosition {
         self.record.position()
     }
 
